from addict import Dict
from django.core.exceptions import PermissionDenied
from django.test import TestCase

from django_scoped_permissions.core import scopes_grant_permissions
from django_scoped_permissions.tests.factories import UserFactory

from django_scoped_permissions.decorators import gql_has_all_scoped_permissions, gql_has_scoped_permissions
from django_scoped_permissions.models import ScopedPermission
from django_scoped_permissions.tests.factories import PetFactory


class TestScopesGrantPermission(TestCase):
    def test__equal_scopes_match__returns_true(self):
        scopes = ["scope1:action"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), True)

    def test__scope_contained_in_required__returns_true(self):
        scopes = ["scope1"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), True)

    def test__no_matching_scope__returns_false(self):
        scopes = ["scope3"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), False)

    def test__scoped_has_negation__invalidates_and_returns_false(self):
        scopes = ["scope1", "-scope1:action"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), False)

    def test__scoped_has_high_level_negation__invalidates_and_returns_false(self):
        scopes = ["scope1:action", "-scope1"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), False)

    def test__scoped_has_exact_operator__returns_true_on_exact_match(self):
        scopes = ["=scope1:action"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), True)

    def test__scoped_has_exact_operator__returns_false_on_exact_match(self):
        scopes = ["=scope1"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), False)

    def test__scoped_has_exact_operator_and_negation__returns_false_on_exact_match(
        self,
    ):
        scopes = ["-=scope1:action"]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), False)

    def test__scoped_has_exact_operator_and_negation__returns_true_on_hierarchical_match(
        self,
    ):
        scopes = [
            "scope1:action",
            "-=scope1",
        ]
        required_scopes = ["scope2", "scope1:action"]
        self.assertEqual(scopes_grant_permissions(required_scopes, scopes), True)


class TestHasScopedPermissionsMixin(TestCase):
    def get_scopes__no_scopes__returns_empty_array(self):
        user = UserFactory.create()

        # Users automatically get the user:1 scope
        self.assertListEqual(
            user.get_scopes(), ["user:1"]
        )

    def test_get_scopes__simple_scopes__returns_array(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope")
        )
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope:deep")
        )

        # Users automatically get the user:1 scope
        self.assertListEqual(
            user.get_scopes(), ["simple:scope", "simple:scope:deep", "user:1"]
        )

    def test_get_scopes__exclude_scopes__appends_minus_signs(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope", exclude=True)
        )
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope:deep", exclude=True)
        )

        # Users automatically get the user:1 scope
        self.assertListEqual(
            user.get_scopes(), ["-simple:scope", "-simple:scope:deep", "user:1"]
        )

    def test_get_scopes__exact_scopes__appends_equal_sign(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope", exact=True)
        )
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope:deep", exact=True)
        )

        # Users automatically get the user:1 scope
        self.assertListEqual(
            user.get_scopes(), ["=simple:scope", "=simple:scope:deep", "user:1"]
        )

    def test_get_scopes__exact_and_negation_scopes__appends_equal_and_minus_sign(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope", exclude=True, exact=True)
        )
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope:deep", exclude=True, exact=True)
        )

        # Users automatically get the user:1 scope
        self.assertListEqual(
            user.get_scopes(), ["-=simple:scope", "-=simple:scope:deep", "user:1"]
        )

class TestScopedModelMixin(TestCase):
    def test_has_permission__user_has_exclude_on_specific_top_level_scope__does_not_cascade(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="user:1:update", exclude=True, exact=True)
        )

        pet = PetFactory.create(user=user)
        self.assertTrue(pet.has_permission(user, "update"))

    def test_has_permission__user_has_exclude_on_specific_top_level_scope_without_action__does_not_cascade(self):
        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="user:1", exclude=True, exact=True)
        )

        pet = PetFactory.create(user=user)
        self.assertTrue(pet.has_permission(user))


class TestGqlHasScopedPermissions(TestCase):

    def test__user_has_permission__succeeds(self):
        @gql_has_scoped_permissions("simple:scope")
        def wrapper_method(cls, info, *args, **kwargs):
            return True

        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope")
        )

        info = Dict(context=Dict(user=user))

        result = wrapper_method(None, info)
        self.assertTrue(result)


    def test__user_does_not_have_permission__fails(self):
        @gql_has_scoped_permissions("simple:scope")
        def wrapper_method(cls, info, *args, **kwargs):
            return True

        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope:nested")
        )

        info = Dict(context=Dict(user=user))

        with self.assertRaises(PermissionDenied):
            wrapper_method(None, info)


class TestGqlHasAllScopedPermissions(TestCase):
    def test__user_has_permission__succeeds(self):
        @gql_has_all_scoped_permissions("simple:scope", "simple:scope:nested", "other:scope")
        def wrapper_method(cls, info, *args, **kwargs):
            return True

        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope")
        )
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="other")
        )

        info = Dict(context=Dict(user=user))

        result = wrapper_method(None, info)
        self.assertTrue(result)


    def test__user_does_not_have_permission__fails(self):
        @gql_has_all_scoped_permissions("simple:scope", "simple:scope:nested", "other:scope")
        def wrapper_method(cls, info, *args, **kwargs):
            return True

        user = UserFactory.create()
        user.scoped_permissions.add(
            ScopedPermission.objects.create(scope="simple:scope")
        )

        info = Dict(context=Dict(user=user))

        with self.assertRaises(PermissionDenied):
            wrapper_method(None, info)



from django.test import TestCase

from django_scoped_permissions.graphql_util import expand_scopes, expand_scopes_from_context


class TestExpandScopes(TestCase):
    def test__no_expansions__return_scopes(self):
        scopes = ["some:scope", "another:scope"]

        self.assertListEqual(scopes, expand_scopes(scopes))
        self.assertListEqual(scopes, expand_scopes(scopes, {}))

    def test__single_expansion_array__returns_interpolated_expansion(self):
        scopes = ["some:{scope}", "another:scope"]

        self.assertListEqual(
            sorted(expand_scopes(scopes, {"scope": ["epic"]})), ["another:scope", "some:epic"]
        )

    def test__multiple_expansion_arrays__returns_interpolated_expansion(self):
        scopes = ["some:{scope}", "another:scope"]

        self.assertListEqual(
            sorted(expand_scopes(scopes, {"scope": ["epic", "epic2"]})),
            sorted(["another:scope", "some:epic", "some:epic2"]),
        )


class TestExpandScopesFromContext(TestCase):
    def test__expand_with_various_objects__succeeds(self):
        scopes = [
            "some:{scope}",
            "some:{deep.context.variable}",
            "some:{class.property}",
        ]

        class SomeClass:
            def __init__(self):
                self.property = "Hi there"

        self.assertListEqual(
            sorted(
                expand_scopes_from_context(
                    scopes,
                    {
                        "scope": 2,
                        "deep": {"context": {"variable": 5}},
                        "class": SomeClass(),
                    },
                )
            ),
            sorted(["some:2", "some:5", "some:Hi there"]),
        )

    def test__expand_with_arrays__calculates_the_products(self):
        scopes = [
            "some:{scope}",
            "some:{deep.context.variable}",
            "some:{class.property}",
        ]

        class SomeClass:
            def __init__(self):
                self.property = "Hi there"

        self.assertListEqual(
            sorted(
                expand_scopes_from_context(
                    scopes,
                    {
                        "scope": [2, 3],
                        "deep": {"context": {"variable": ["result-1", "result-2"]}},
                        "class": SomeClass(),
                    },
                )
            ),
            sorted(
                ["some:2", "some:3", "some:result-1", "some:result-2", "some:Hi there"]
            ),
        )



import factory

from django_scoped_permissions.tests.models import User, Pet


class UserFactory(factory.DjangoModelFactory):
    class Meta:
        model = User

    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    username = factory.Sequence(lambda n: "username%d" % n)
    email = factory.Sequence(lambda n: "user-%d@ursolutions.no" % n)


class PetFactory(factory.DjangoModelFactory):
    class Meta:
        model = Pet

    user = factory.SubFactory(UserFactory)
    name = "Pet"
    age = 10



from django.contrib import admin

from django_scoped_permissions.models import ScopedPermission


@admin.register(ScopedPermission)
class ScopedPermissionAdmin(admin.ModelAdmin):
    list_display = (
        "__str__",
        "exact",
        "exclude",
    )



from typing import Optional

from django.db import models
from django.db.models import Value, F, Case, When
from django.db.models.functions import Concat
from django.utils.functional import cached_property

from django_scoped_permissions.core import any_scope_matches, scopes_grant_permissions


class ScopedPermission(models.Model):
    class Meta:
        unique_together = (("scope", "exclude", "exact"),)

    scope = models.TextField(blank=False)
    exclude = models.BooleanField(
        default=False,
        help_text="Whether this should be an exclusive permission, meaning that if scope is 'user:update' and exclude is True then users with this usertype will not be able to update users, even their own.",
    )
    exact = models.BooleanField(
        default=False,
        help_text="If checked, the permission needs an exact match to count. In other words, it does not work recursively as standard scoped permissions.",
    )

    def get_scope_parts(self):
        return self.scope.split(":")

    def __str__(self):
        prefix = "-" if self.exclude else ""
        prefix += "=" if self.exact else ""
        return prefix + self.scope


class HasScopedPermissionsMixin(models.Model):
    class Meta:
        abstract = True

    scoped_permissions = models.ManyToManyField(ScopedPermission, blank=True,)

    @cached_property
    def resolved_scopes(self):
        scopes = self.scoped_permissions
        scopes = scopes.annotate(
            parsed_scope=Concat(
                Case(When(exclude=True, then=Value("-")), default=Value("")),
                Case(When(exact=True, then=Value("=")), default=Value("")),
                F("scope"),
            )
        )

        specific_scopes = list(scopes.values_list("parsed_scope", flat=True))

        return specific_scopes

    def get_scopes(self):
        return self.resolved_scopes

    def has_scoped_permissions(self, *required_scopes):
        return self.has_any_scoped_permissions(*required_scopes)

    def has_any_scoped_permissions(self, *required_scopes):
        scopes = self.get_scopes()

        return scopes_grant_permissions(required_scopes, scopes)

    def has_all_scoped_permissions(self, *required_scopes):
        scopes = self.get_scopes()

        for scope in required_scopes:
            if not any_scope_matches([scope], scopes):
                return False

        return True

    def has_create_permission(self, model_name: str, action: str = "create"):
        scopes = self.get_scopes()
        return f"-{model_name}:{action}" not in scopes


class ScopedModelMixin:
    def get_base_scopes(self):
        return []

    def has_permission(
            self, user: HasScopedPermissionsMixin, action: Optional[str] = None
    ):
        if getattr(user, "is_superuser", False):
            return True

        user_scopes = user.get_scopes()
        base_scopes = self.get_base_scopes()

        return scopes_grant_permissions(base_scopes, user_scopes)


class ScopedModel(models.Model, ScopedModelMixin):
    class Meta:
        abstract = True



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="django-scoped-permissions",
    version="0.0.4",
    author="Tormod Haugland, Magnus Buvarp",
    author_email="tormod.haugland@gmail.com, magnus.buvarp@gmail.com",
    description="Scoped permission system for Django",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ur-Solutions/django_scoped_permissions",
    packages=setuptools.find_packages(),
    install_requires=[
        "django>=2.0.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)
